<?php

declare(strict_types=1);
namespace Drupal\translation_bliss\Drush\Commands\Extensions;

use Consolidation\AnnotatedCommand\CommandData;
use Consolidation\AnnotatedCommand\Hooks\HookManager;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Config\TypedConfigManagerInterface;
use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\Core\TypedData\TraversableTypedDataInterface;
use Drupal\Core\TypedData\TypedDataInterface;
use Drupal\translation_bliss\Drush\Commands\TranslationBlissPotxCommands;
use Drush\Attributes as CLI;
use Drush\Commands\AutowireTrait;
use Drush\Commands\DrushCommands;

final class TranslationBlissConfigPotxCommands extends DrushCommands {

  use AutowireTrait;

  const DoNotExtractOption = 'no-config-sync';

  protected bool $doNotExtract;

  public function __construct(
    private readonly ConfigFactoryInterface $configFactory,
    private readonly TypedConfigManagerInterface $typedConfigManager,
    private readonly ModuleHandlerInterface $moduleHandler,
  ) {
    parent::__construct();
  }

  #[CLI\Hook(type: HookManager::OPTION_HOOK, target: TranslationBlissPotxCommands::PotxxCommandName)]
  #[CLI\Option(name: self::DoNotExtractOption, description: 'Do not extract config/sync values.')]
  public function optionsPotxX($options = [self::DoNotExtractOption => FALSE]): void {}

  #[CLI\Hook(type: HookManager::ARGUMENT_VALIDATOR, target: TranslationBlissPotxCommands::PotxxCommandName)]
  public function readOptions(CommandData $commandData): void {
    $this->doNotExtract = $commandData->input()->getOption(self::DoNotExtractOption) ?? FALSE;
  }

  #[CLI\Hook(type: HookManager::ON_EVENT, target: TranslationBlissPotxCommands::DrushHookFinishProcessing)]
  public function hookPotxFinishProcessing(): void {
    if ($this->doNotExtract) {
      $this->output()->writeln('Skipping config extraction.');
      return;
    }
    $this->output()->writeln('Extracting config.');

    // Parsing config via _potx_process_module_schemas works out badly, as
    // there is no simple way to determine the providing module and load its
    // schema.
    // So load runtime config, but use config/sync directory to hint file.
    $dirPath = \Drupal\Core\Site\Settings::get('config_sync_directory');

    $configNames = $this->configFactory->listAll();
    foreach ($configNames as $configName) {
      $this->output()->writeln("Processing config $configName");
      $this->processTypedConfig($this->typedConfigManager->get($configName), "$dirPath/$configName.yml");
    }
  }

  private function processTypedConfig(TypedDataInterface $element, string $fileName, array $keys = []): void {
    /** Adapted from @see \Drupal\locale\LocaleConfigManager::getTranslatableData */
    if ($element instanceof TraversableTypedDataInterface) {
      foreach ($element as $key => $property) {
        $this->processTypedConfig($property, $fileName, array_merge($keys, [$key]));
      }
    }
    else {
      $value = $element->getValue();
      $definition = $element->getDataDefinition();
      if (!empty($definition['translatable']) && $value !== '' && $value !== NULL) {
        $context = $definition['translation context'] ?? NULL;
        _potx_save_string($value, $context, $fileName);
      }
    }
  }

  /**
   * @param list<string> $sourceDirs
   */
  #[\Drush\Attributes\Hook(type: HookManager::ON_EVENT, target: TranslationBlissPotxCommands::DrushHookLocalInit)]
  public function hookLocalInit(array $sourceDirs): void {
    // Nothing to do. Just to show off the API:
  }

  /**
   * @param list<string> $sourceDirs
   *
   * @return list<string>
   *   A list of files to process.
   */
  #[\Drush\Attributes\Hook(type: HookManager::ON_EVENT, target: TranslationBlissPotxCommands::DrushHookExploreDirs)]
  public function hookExploreDirs(array $sourceDirs): array {
    // Nothing to do. Just to show off the API:
    return $sourceDirs;
  }

  /**
   * @param list<string> $files
   */
  #[\Drush\Attributes\Hook(type: HookManager::ON_EVENT, target: TranslationBlissPotxCommands::DrushHookProcessFiles)]
  public function hookProcessFiles(array $files): void {
    // Nothing to do. Just to show off the API:
  }


}
