<?php

declare(strict_types=1);
namespace Drupal\translation_bliss\Drush\Commands;

use Consolidation\AnnotatedCommand\CommandData;
use Consolidation\AnnotatedCommand\Events\CustomEventAwareInterface;
use Consolidation\AnnotatedCommand\Events\CustomEventAwareTrait;
use Consolidation\AnnotatedCommand\Hooks\HookManager;
use Consolidation\OutputFormatters\StructuredData\RowsOfFields;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\Core\File\FileSystemInterface;
use Drupal\Core\Language\LanguageManagerInterface;
use Drupal\translation_bliss\Drush\Commands\Utility\PotxResult;
use Drupal\translation_bliss\Drush\Commands\Utility\TmpCwd;
use Drush\Attributes as CLI;
use Drush\Commands\AutowireTrait;
use Drush\Commands\DrushCommands;
use Symfony\Component\Console\Input\InputOption;

/**
 * Class PotxXCommands.
 *
 * Potx commands for site builders.
 */
class TranslationBlissPotxCommands extends DrushCommands implements CustomEventAwareInterface {

  use CustomEventAwareTrait;
  use AutowireTrait;

  public const PotxxCommandName = 'translation-bliss:potx';
  public const TargetFileBaseName = 'general';

  // Do not use POTX_API_CURRENT, it breaks drush discovery.
  public const PotxApi = 8;
  public const PotxSaveCallback = '_potx_save_string';
  public const PotxVersionCallback = '_potx_save_version';


  const DrushHookLocalInit = 'potxx_local_init';
  const DrushHookExploreDirs = 'potxx_explore_dirs';
  const DrushHookProcessFiles = 'potxx_process_files';
  const DrushHookFinishProcessing = 'potxx_finish_processing';

  const DefaultSourceDirs = 'modules/custom';

  const DefaultTargetDir = '../translations/custom';

  /**
   * @var list<string>
   */
  protected array $sourceDirs;

  protected string $targetDir;

  public function __construct(
    protected readonly ModuleHandlerInterface $moduleHandler,
    protected readonly LanguageManagerInterface $languageManager,
    protected readonly ConfigFactoryInterface $configFactory,
    protected readonly FileSystemInterface $fileSystem,
  ) {
    parent::__construct();
  }

  #[CLI\Hook(type: HookManager::ARGUMENT_VALIDATOR, target: self::PotxxCommandName)]
  public function validateExtract(CommandData $commandData): void {
    $options = $commandData->options();

    /** @var string[] $sourceDirs */
    $this->sourceDirs = explode(',', $options['source'] ?: static::DefaultSourceDirs);
    foreach ($this->sourceDirs as &$sourceDir) {
      $this->normalizeDir($sourceDir);
      $this->validateDir($sourceDir);
    }

    /** @var string $target */
    $this->targetDir = $options['target'] ?: self::DefaultTargetDir;
    $this->normalizeDir($this->targetDir);
    $this->validateDir($this->targetDir, createIfNeeded: TRUE);
  }

  private function normalizeDir(string &$dir): void {
    $dir = rtrim($dir, '/');
  }

  private function validateDir(string $dir, $createIfNeeded = FALSE): void {
    if ($createIfNeeded && !is_dir($dir)) {
      $this->fileSystem->prepareDirectory($dir, FileSystemInterface::CREATE_DIRECTORY | FileSystemInterface::MODIFY_PERMISSIONS);
    }
    if (!is_dir($dir)) {
      $cwd = getcwd();
      throw new \UnexpectedValueException("Not a directory: '$dir', cwd='$cwd'");
    }
  }

  #[CLI\Hook(type: HookManager::ARGUMENT_VALIDATOR, target: TranslationBlissPotxCommands::PotxxCommandName)]
  public function validatePotxInstalled(CommandData $commandData): void {
    if (!$this->moduleHandler->moduleExists('potx')) {
      throw new \Exception('Please install the potx module required for this command.');
    }
  }

  /**
   * Extract translatable strings for site builders.
   */
  #[CLI\Command(name: self::PotxxCommandName, aliases: ['potxx'])]
  #[CLI\Usage(name: 'potxx', description: 'Translation template extraction optimized for site owners.')]
  #[CLI\Option(name: 'source', description: 'One or more source folders. Defaults to "modules/custom".')]
  #[CLI\Option(name: 'target', description: 'Target folder. Defaults to "../translations/custom".')]
  #[CLI\FieldLabels(labels: [
    'files' => 'Files',
    'strings' => 'Strings',
    'errors' => 'Errors',
  ])]
  public function extractToFiles(array $options = [
    'source' => InputOption::VALUE_REQUIRED,
    'target' => InputOption::VALUE_REQUIRED,
  ]): RowsOfFields {
    // Extract.
    $result = $this->extract($this->sourceDirs);
    $rows[] = [
      'files' => count($result->files),
      'strings' => count($result->strings),
      'errors' => count($result->errors),
    ];

    // Build general.pot
    $this->buildFile(NULL, $renames);

    // Build general.XX.po
    foreach ($this->languageManager->getLanguages() as $language) {
      if ($language->getId() === 'en') {
        if (!$this->configFactory->get('locale.settings')->get('translate_english')) {
          continue;
        }
      }
      $this->buildFile($language->getId(), $renames);
    }

    // Write and rename.
    $this->writeFiles($renames);

    // Print results.
    if (!empty($result->errors)) {
      // Fool potx.
      $logError = [$this->logger(), 'error'];
      $this->output()->writeln(dt('Errors'));
      foreach (array_unique($result->errors) as $error) {
        $logError($error);
      }
    }
    $this->output()->writeln('');
    return new RowsOfFields($rows);
  }

  protected function extract(
    array $sourceDirs,
  ): PotxResult {
    // Include library.
    $this->moduleHandler->loadInclude('potx', 'inc');
    $this->moduleHandler->loadInclude('potx', 'inc', 'potx.local');

    $files = [];

    // Silence error message reporting. Messages will be reported by at the end.
    potx_status('set', POTX_STATUS_SILENT);

    foreach ($sourceDirs as $sourceDir) {
      potx_local_init($sourceDir);
    }
    foreach ($this->getCustomEventHandlers(self::DrushHookLocalInit) as $handler) {
      $handler($sourceDirs);
    }

    foreach ($sourceDirs as $sourceDir) {
      $files = array_merge($files, _potx_explore_dir($sourceDir, '*', self::PotxApi, TRUE));
    }
    foreach ($this->getCustomEventHandlers(self::DrushHookExploreDirs) as $handler) {
      $hookFiles = $handler($sourceDirs);
      $files = array_merge($files, $hookFiles);
    }

    foreach ($files as $file) {
      $this->output()->writeln("Processing $file...");
      _potx_process_file($file, 0, self::PotxSaveCallback, self::PotxVersionCallback, self::PotxApi);
    }
    foreach ($this->getCustomEventHandlers(self::DrushHookProcessFiles) as $handler) {
      $handler($files);
    }

    potx_finish_processing(self::PotxSaveCallback, self::PotxApi);
    foreach ($this->getCustomEventHandlers(self::DrushHookFinishProcessing) as $handler) {
      $handler();
    }

    // Get errors, if any.
    $errors = potx_status('get', TRUE);

    // Get saved strings.
    /** @var string[] $strings */
    $strings = _potx_save_string(NULL, NULL, NULL, 0, POTX_STRING_RUNTIME);

    return new PotxResult($files, $strings, $errors);
  }

  protected function buildFile(
    ?string $langcode = NULL,
    array &$renames = NULL,
  ): void {
    // Potx is very limited wrt generated file name, so use temporary name.
    $temporaryName = $langcode ? "_general_$langcode" : '_general';

    _potx_build_files(
      POTX_STRING_RUNTIME,
      POTX_BUILD_SINGLE,
      $temporaryName,
      '_potx_save_string',
      '_potx_save_version',
      '_potx_get_header',
      $langcode,
      $langcode,
      self::PotxApi);

    $targetPath = $langcode ?
      sprintf('%s/%s.%s.po', $this->targetDir, self::TargetFileBaseName, $langcode) :
      sprintf('%s/%s.pot', $this->targetDir, self::TargetFileBaseName);
    // _potx_build_files always appends .pot.
    $renames["$this->targetDir/$temporaryName.pot"] = $targetPath;
  }

  protected function writeFiles(
    array $renames,
  ): void {
    $tmpCwd = TmpCwd::create($this->targetDir)
      ?? throw new \RuntimeException("Can not change to dir $this->targetDir");

    _potx_write_files();

    unset($tmpCwd);

    foreach ($renames as $from => $to) {
      rename($from, $to);
    }
  }

}
